@@ -22,9 +22,11 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
-   $Date: 2013-08-21 11:45:19 +0100 (Wed, 21 Aug 2013) $ $Revision: 36999 $
+   $Date$ $Revision$
    */ 
 
+#include <inttypes.h>
+
 #import	<Foundation/NSArray.h>
 #import	<Foundation/NSAutoreleasePool.h>
 #import	<Foundation/NSData.h>
@@ -37,9 +39,9 @@
 #import	<Foundation/NSLock.h>
 #import	<Foundation/NSMapTable.h>
 #import	<Foundation/NSNotification.h>
-#import	<Foundation/NSSet.h>
 #import	<Foundation/NSString.h>
 #import	<Foundation/NSThread.h>
+#import	<Foundation/NSUserDefaults.h>
 #import	<Foundation/NSValue.h>
 
 #import	"GSCache.h"
@@ -50,11 +52,13 @@
 #if (MAC_OS_X_VERSION_MAX_ALLOWED <= MAC_OS_X_VERSION_10_4)
 #define class_getInstanceSize(isa)  ((struct objc_class *)isa)->instance_size
 #endif
+
+#import "NSObject+GSExtensions.h"
+
 #endif
 
-@interface	GSCache (Threading)
-+ (void) _becomeThreaded: (NSNotification*)n;
-- (void) _createLock;
+@interface	GSCache (Private)
+- (void) _useDefaults: (NSNotification*)n;
 @end
 
 @interface	GSCacheItem : NSObject
@@ -88,6 +92,17 @@
   [object release];
   [super dealloc];
 }
+- (NSUInteger) sizeInBytesExcluding: (NSHashTable*)exclude
+{
+  NSUInteger    bytes = [super sizeInBytesExcluding: exclude];
+
+  if (bytes > 0)
+    {
+      bytes += [key sizeInBytesExcluding: exclude];
+      bytes += [object sizeInBytesExcluding: exclude];
+    }
+  return bytes;
+}
 @end
 
 
@@ -111,8 +126,9 @@ typedef struct {
   NSMapTable	*contents;
   GSCacheItem	*first;
   NSString	*name;
-  NSMutableSet	*exclude;
+  NSHashTable	*exclude;
   NSRecursiveLock	*lock;
+  BOOL		useDefaults;
 } Item;
 #define	my	((Item*)((void*)self + itemOffset))
 
@@ -202,25 +218,7 @@ static void removeItem(GSCacheItem *item, GSCacheItem **first)
   if (allCaches == 0)
     {
       itemOffset = class_getInstanceSize(self);
-      allCaches
-	= NSCreateHashTable(NSNonRetainedObjectHashCallBacks, 0);
-      if ([NSThread isMultiThreaded] == YES)
-	{
-	  [self _becomeThreaded: nil];
-	}
-      else
-	{
-	  /* If and when we become multi-threaded, the +_becomeThreaded:
-	   * method will remove us as an observer and will create a lock
-	   * for the table of all caches, then ask each cache to create
-	   * its own lock.
-	   */
-	  [[NSNotificationCenter defaultCenter]
-	    addObserver: self
-	    selector: @selector(_becomeThreaded:)
-	    name: NSWillBecomeMultiThreadedNotification
-	    object: nil];
-	}
+      allCaches = NSCreateHashTable(NSNonRetainedObjectHashCallBacks, 0);
       GSTickerTimeNow();
     }
 }
@@ -237,6 +235,11 @@ static void removeItem(GSCacheItem *item, GSCacheItem **first)
 
 - (void) dealloc
 {
+  if (my->useDefaults)
+    {
+      [[NSNotificationCenter defaultCenter] removeObserver: self
+	name: NSUserDefaultsDidChangeNotification object: nil];
+    }
   [allCachesLock lock];
   NSHashRemove(allCaches, (void*)self);
   [allCachesLock unlock];
@@ -287,10 +290,7 @@ static void removeItem(GSCacheItem *item, GSCacheItem **first)
 {
   if (nil != (self = [super init]))
     {
-      if ([NSThread isMultiThreaded] == YES)
-	{
-	  [self _createLock];
-	}
+      my->lock = [NSRecursiveLock new];
       my->contents = NSCreateMapTable(NSObjectMapKeyCallBacks,
 	NSObjectMapValueCallBacks, 0);
       [allCachesLock lock];
@@ -522,6 +522,46 @@ static void removeItem(GSCacheItem *item, GSCacheItem **first)
     }
 }
 
+- (id) refreshObject: (id)anObject
+              forKey: (id)aKey
+            lifetime: (unsigned)lifetime
+{
+  id		object;
+  GSCacheItem	*item;
+
+  [my->lock lock];
+  item = (GSCacheItem*)NSMapGet(my->contents, aKey);
+  if (item == nil)
+    {
+      if (nil != anObject)
+        {
+          [self setObject: anObject
+                   forKey: aKey
+                 lifetime: lifetime];
+        }
+      [my->lock unlock];
+      return anObject;
+    }
+
+  if (nil != anObject && NO == [anObject isEqual: item->object])
+    {
+      object = [anObject retain];
+      [item->object release];
+      item->object = object;
+    }
+  if (lifetime > 0)
+    {
+      unsigned	tick = GSTickerTimeTick();
+
+      item->when = tick + lifetime;
+      item->warn = tick + lifetime / 2;
+    }
+  item->life = lifetime;
+  object = [[item->object retain] autorelease];
+  [my->lock unlock];
+  return object;
+}
+
 - (void) setDelegate: (id)anObject
 {
   [my->lock lock];
@@ -553,17 +593,41 @@ static void removeItem(GSCacheItem *item, GSCacheItem **first)
 
 - (void) setLifetime: (unsigned)max
 {
+  [my->lock lock];
+  if (YES == my->useDefaults)
+    {
+      NSUserDefaults	*defs = [NSUserDefaults standardUserDefaults];
+      NSString          *n = (nil == my->name) ? @"" : my->name;
+      NSString          *k = [@"GSCacheLifetime" stringByAppendingString: n];
+
+      if (nil != [defs objectForKey: k])
+        {
+          max = (unsigned) [defs integerForKey: k];
+        }
+    }
   my->lifetime = max;
+  [my->lock unlock];
 }
 
 - (void) setMaxObjects: (unsigned)max
 {
   [my->lock lock];
+  if (YES == my->useDefaults)
+    {
+      NSUserDefaults	*defs = [NSUserDefaults standardUserDefaults];
+      NSString          *n = (nil == my->name) ? @"" : my->name;
+      NSString          *k = [@"GSCacheMaxObjects" stringByAppendingString: n];
+
+      if (nil != [defs objectForKey: k])
+        {
+          max = (unsigned) [defs integerForKey: k];
+        }
+    }
   my->maxObjects = max;
   if (my->currentObjects > my->maxObjects)
     {
       [self shrinkObjects: my->maxObjects
-		  andSize: my->maxSize];
+                  andSize: my->maxSize];
     }
   [my->lock unlock];
 }
@@ -571,6 +635,17 @@ static void removeItem(GSCacheItem *item, GSCacheItem **first)
 - (void) setMaxSize: (NSUInteger)max
 {
   [my->lock lock];
+  if (YES == my->useDefaults)
+    {
+      NSUserDefaults	*defs = [NSUserDefaults standardUserDefaults];
+      NSString          *n = (nil == my->name) ? @"" : my->name;
+      NSString          *k = [@"GSCacheMaxSize" stringByAppendingString: n];
+
+      if (nil != [defs objectForKey: k])
+        {
+          max = (NSUInteger) [defs integerForKey: k];
+        }
+    }
   if (max > 0 && my->maxSize == 0)
     {
       NSMapEnumerator	e = NSEnumerateMapTable(my->contents);
@@ -578,30 +653,31 @@ static void removeItem(GSCacheItem *item, GSCacheItem **first)
       id		k;
       NSUInteger	size = 0;
 
-      if (my->exclude == nil)
-	{
-	  my->exclude = [NSMutableSet new];
-	}
+      if (nil == my->exclude)
+        {
+          my->exclude
+            = NSCreateHashTable(NSNonOwnedPointerHashCallBacks, 0);
+        }
       while (NSNextMapEnumeratorPair(&e, (void**)&k, (void**)&i) != 0)
-	{
-	  if (i->size == 0)
-	    {
-	      [my->exclude removeAllObjects];
-	      i->size = [i->object sizeInBytes: my->exclude];
-	    }
-	  if (i->size > max)
-	    {
-	      /*
-	       * Item in cache is too big for new size limit ...
-	       * Remove it.
-	       */
-	      removeItem(i, &my->first);
-	      NSMapRemove(my->contents, (void*)i->key);
-	      my->currentObjects--;
-	      continue;
-	    }
-	  size += i->size;
-	}
+        {
+          if (i->size == 0)
+            {
+              i->size = [i->object sizeInBytesExcluding: my->exclude];
+              [my->exclude removeAllObjects];
+            }
+          if (i->size > max)
+            {
+              /*
+               * Item in cache is too big for new size limit ...
+               * Remove it.
+               */
+              removeItem(i, &my->first);
+              NSMapRemove(my->contents, (void*)i->key);
+              my->currentObjects--;
+              continue;
+            }
+          size += i->size;
+        }
       NSEndMapTableEnumeration(&e);
       my->currentSize = size;
     }
@@ -613,18 +689,44 @@ static void removeItem(GSCacheItem *item, GSCacheItem **first)
   if (my->currentSize > my->maxSize)
     {
       [self shrinkObjects: my->maxObjects
-		  andSize: my->maxSize];
+                  andSize: my->maxSize];
+    }
+  [my->lock unlock];
+}
+
+- (void) setName: (NSString*)name forConfiguration: (BOOL)useDefaults
+{
+  NSString	*c;
+
+  [my->lock lock];
+  c = [name copy];
+  [my->name release];
+  my->name = c;
+  useDefaults = (useDefaults ? YES : NO);	// Make sure this is a real bool
+  if (my->useDefaults != useDefaults)
+    {
+      if (my->useDefaults)
+	{
+	  [[NSNotificationCenter defaultCenter] removeObserver: self
+	    name: NSUserDefaultsDidChangeNotification object: nil];
+	}
+      my->useDefaults = useDefaults;
+      if (my->useDefaults)
+	{
+	  [[NSNotificationCenter defaultCenter]
+	    addObserver: self
+	    selector: @selector(_useDefaults:)
+	    name: NSUserDefaultsDidChangeNotification
+	    object: nil];
+	  [self _useDefaults: nil];
+	}
     }
   [my->lock unlock];
 }
 
 - (void) setName: (NSString*)name
 {
-  [my->lock lock];
-  [name retain];
-  [my->name release];
-  my->name = name;
-  [my->lock unlock];
+  [self setName: name forConfiguration: NO];
 }
 
 - (void) setObject: (id)anObject forKey: (id)aKey
@@ -661,12 +763,13 @@ static void removeItem(GSCacheItem *item, GSCacheItem **first)
     {
       if (maxSize > 0)
 	{
-	  if (my->exclude == nil)
-	    {
-	      my->exclude = [NSMutableSet new];
-	    }
+          if (nil == my->exclude)
+            {
+              my->exclude
+                = NSCreateHashTable(NSNonOwnedPointerHashCallBacks, 0);
+            }
+	  addSize = [anObject sizeInBytesExcluding: my->exclude];
 	  [my->exclude removeAllObjects];
-	  addSize = [anObject sizeInBytes: my->exclude];
 	  if (addSize > maxSize)
 	    {
 	      addObjects = 0;	// Object too big to cache.
@@ -761,197 +864,66 @@ static void removeItem(GSCacheItem *item, GSCacheItem **first)
     }
   [my->lock unlock];
 }
-@end
-@implementation	GSCache (Threading)
-+ (void) _becomeThreaded: (NSNotification*)n
-{
-  NSHashEnumerator	e;
-  GSCache		*c;
 
-  [[NSNotificationCenter defaultCenter] removeObserver: self
-    name: NSWillBecomeMultiThreadedNotification object: nil];
-  allCachesLock = [NSRecursiveLock new];
-  e = NSEnumerateHashTable(allCaches);
-  while ((c = (GSCache*)NSNextHashEnumeratorItem(&e)) != nil)
-    {
-      [c _createLock];
-    }
-  NSEndHashTableEnumeration(&e);
-}
-- (void) _createLock
+- (NSUInteger) sizeInBytesExcluding: (NSHashTable*)exclude
 {
-  my->lock = [NSRecursiveLock new];
-}
-@end
-
-@implementation	NSArray (GSCacheSizeInBytes)
-- (NSUInteger) sizeInBytes: (NSMutableSet*)exclude
-{
-  NSUInteger	size = [super sizeInBytes: exclude];
+  NSUInteger    size;
 
+  [my->lock lock];
+  size = [super sizeInBytesExcluding: exclude];
   if (size > 0)
     {
-      NSUInteger	count = [self count];
+      size += sizeof(Item)
+        + [my->contents sizeInBytesExcluding: exclude]
+        + [my->exclude sizeInBytesExcluding: exclude]
+        + [my->name sizeInBytesExcluding: exclude]
+        + [my->lock sizeInBytesExcluding: exclude];
+    }
+  [my->lock unlock];
+  return size;
+}
 
-      size += count*sizeof(void*);
-      while (count-- > 0)
+@end
+@implementation	GSCache (Private)
+- (void) _useDefaults: (NSNotification*)n
+{
+  NSUserDefaults	*defs = [NSUserDefaults standardUserDefaults];
+  NSString		*conf = (nil == my->name ? @"" : my->name);
+  NSString		*key;
+
+  [my->lock lock];
+  NS_DURING
+    {
+      if (YES == my->useDefaults)
 	{
-	  size += [[self objectAtIndex: count] sizeInBytes: exclude];
-	}
-    }
-  return size;
-}
-@end
-
-@implementation	NSData (GSCacheSizeInBytes)
-- (NSUInteger) sizeInBytes: (NSMutableSet*)exclude
-{
-  NSUInteger	size = [super sizeInBytes: exclude];
-
-  if (size > 0)
-    {
-      size += [self length];
-    }
-  return size;
-}
-@end
-
-@implementation	NSDictionary (GSCacheSizeInBytes)
-- (NSUInteger) sizeInBytes: (NSMutableSet*)exclude
-{
-  NSUInteger	size = [super sizeInBytes: exclude];
-
-  if (size > 0)
-    {
-      NSUInteger	count = [self count];
-
-      size += 3 * sizeof(void*) * count;
-      if (count > 0)
-        {
-	  NSAutoreleasePool	*pool = [NSAutoreleasePool new];
-	  NSEnumerator		*enumerator = [self keyEnumerator];
-	  NSObject		*k;
-
-	  while ((k = [enumerator nextObject]) != nil)
+	  my->useDefaults = NO;
+	  key = [@"GSCacheLifetime" stringByAppendingString: conf];
+	  if (nil != [defs objectForKey: key])
 	    {
-	      NSObject	*o = [self objectForKey: k];
-
-	      size += [k sizeInBytes: exclude] + [o sizeInBytes: exclude];
+	      [self setLifetime: (unsigned)[defs integerForKey: key]];
 	    }
-	  [pool release];
-	}
-    }
-  return size;
-}
-@end
-
-@implementation	NSObject (GSCacheSizeInBytes)
-- (NSUInteger) sizeInBytes: (NSMutableSet*)exclude
-{
-  if ([exclude member: self] != nil)
-    {
-      return 0;
-    }
-  [exclude addObject: self];
-
-  return class_getInstanceSize(object_getClass(self)); 
-}
-@end
-
-@implementation	NSSet (GSCacheSizeInBytes)
-- (NSUInteger) sizeInBytes: (NSMutableSet*)exclude
-{
-  NSUInteger	size = [super sizeInBytes: exclude];
-
-  if (size > 0)
-    {
-      NSUInteger	count = [self count];
-
-      size += 3 * sizeof(void*) * count;
-      if (count > 0)
-        {
-	  NSAutoreleasePool	*pool = [NSAutoreleasePool new];
-	  NSEnumerator		*enumerator = [self objectEnumerator];
-	  NSObject		*o;
-
-	  while ((o = [enumerator nextObject]) != nil)
+	  key = [@"GSCacheMaxObjects" stringByAppendingString: conf];
+	  if (nil != [defs objectForKey: key])
 	    {
-	      size += [o sizeInBytes: exclude];
+	      [self setMaxObjects: (unsigned)[defs integerForKey: key]];
 	    }
-	  [pool release];
+	  key = [@"GSCacheMaxSize" stringByAppendingString: conf];
+	  if (nil != [defs objectForKey: key])
+	    {
+	      [self setMaxSize: (NSUInteger)[defs integerForKey: key]];
+	    }
+	  my->useDefaults = YES;
 	}
+      [my->lock unlock];
     }
-  return size;
-}
-@end
-
-@implementation	NSString (GSCacheSizeInBytes)
-- (NSUInteger) sizeInBytes: (NSMutableSet*)exclude
-{
-  if ([exclude member: self] != nil)
+  NS_HANDLER
     {
-      return 0;
-    }
-  else
-    {
-      return [super sizeInBytes: exclude] + sizeof(unichar) * [self length];
+      my->useDefaults = YES;
+      [my->lock unlock];
+      [localException raise];
     }
+  NS_ENDHANDLER
 }
 @end
 
-#if	defined(GNUSTEP_BASE_LIBRARY)
-
-// FIXME ... this should be moved to base additions.
-#import	<GNUstepBase/GSMime.h>
-
-#if	GS_NONFRAGILE
-/* When using the non-fragile ABI, base does not expose the mime class
- * instance variables, so we use the raw data size as a rough approximation.
- * The actual memory usage will be larger than this of course.
- */
-@implementation	GSMimeDocument (GSCacheSizeInBytes)
-- (NSUInteger) sizeInBytes: (NSMutableSet*)exclude
-{
-  return [[self rawMimeData] sizeInBytes: exclude];
-}
-@end
-
-@implementation	GSMimeHeader (GSCacheSizeInBytes)
-- (NSUInteger) sizeInBytes: (NSMutableSet*)exclude
-{
-  return [[self rawMimeData] sizeInBytes: exclude];
-}
-@end
-#else
-@implementation	GSMimeDocument (GSCacheSizeInBytes)
-- (NSUInteger) sizeInBytes: (NSMutableSet*)exclude
-{
-  NSUInteger	size = [super sizeInBytes: exclude];
-
-  if (size > 0)
-    {
-      size += [content sizeInBytes: exclude] + [headers sizeInBytes: exclude];
-    }
-  return size;
-}
-@end
-
-@implementation	GSMimeHeader (GSCacheSizeInBytes)
-- (NSUInteger) sizeInBytes: (NSMutableSet*)exclude
-{
-  NSUInteger	size = [super sizeInBytes: exclude];
-
-  if (size > 0)
-    {
-      size += [name sizeInBytes: exclude]
-        + [value sizeInBytes: exclude]
-        + [objects sizeInBytes: exclude]
-        + [params sizeInBytes: exclude];
-    }
-  return size;
-}
-@end
-#endif
-
-#endif
 
