@@ -22,7 +22,7 @@
    */
 #import "GSFIFO.h"
 #import <Foundation/NSArray.h>
-#import <Foundation/NSDate.h>
+#import <Foundation/NSException.h>
 #import <Foundation/NSLock.h>
 #import <Foundation/NSMapTable.h>
 #import <Foundation/NSString.h>
@@ -31,6 +31,12 @@
 #import <Foundation/NSValue.h>
 #import <Foundation/NSZone.h>
 
+#if !defined(GNUSTEP)
+#import "NSObject+GSExtensions.h"
+#endif
+
+#include <inttypes.h>
+
 @implementation	GSFIFO
 
 static NSLock		*classLock = nil;
@@ -126,6 +132,7 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
 - (unsigned) _cooperatingGet: (void**)buf
 		       count: (unsigned)count
 		 shouldBlock: (BOOL)block
+                      before: (NSDate*)before
 {
   NSTimeInterval	ti;
   unsigned		index;
@@ -143,7 +150,7 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
 	}
 
       START
-      if (0 == timeout)
+      if ((0 == timeout) && (before == nil))
 	{
 	  while (_head - _tail == 0)
 	    {
@@ -152,23 +159,49 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
 	}
       else
 	{
-	  NSDate	*d;
+	  NSDate	*d = nil;
+          NSDate        *effective;
 
-	  d = [[NSDateClass alloc]
-	    initWithTimeIntervalSinceNow: timeout / 1000.0f];
+          [before retain];
+          if (timeout != 0)
+            {
+              d = [[NSDateClass alloc]
+               initWithTimeIntervalSinceNow: timeout / 1000.0f];
+            }
+          if ((d != nil) && (before != nil))
+            {
+              effective = [d earlierDate: before];
+            }
+          else if (d != nil)
+            {
+              effective = d;
+            }
+          else
+            {
+              effective = before;
+            }
 	  while (_head - _tail == 0)
 	    {
-	      if (NO == [condition waitUntilDate: d])
+	      if (NO == [condition waitUntilDate: effective])
 		{
 		  [d release];
+                  [before release];
 		  ENDGET
 		  [condition broadcast];
 		  [condition unlock];
-		  [NSException raise: NSGenericException
-			      format: @"Timeout waiting for new data in FIFO"];
+                  if (before != effective)
+                    {
+                      [NSException raise: NSGenericException
+                        format: @"Timeout waiting for new data in FIFO"];
+                    }
+                  else
+                    {
+                      return 0;
+                    }
 		}
 	    }
 	  [d release];
+          [before release];
 	  ENDGET
 	}
     }
@@ -199,6 +232,64 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
   return index;
 }
 
+- (void*) _cooperatingPeek
+{
+  [condition lock];
+  if ((_head - _tail) == 0)
+    {
+      // We do not need to signal the condition because
+      // nothing about the qeuue did change
+      [condition unlock];
+      return NULL;
+    }
+  void *ptr = _items[_tail % _capacity];
+  [condition unlock];
+  return ptr;
+}
+
+- (id) _cooperatingPeekObject
+{
+  [condition lock];
+  if ((_head - _tail) == 0)
+    {
+      // We do not need to signal the condition because
+      // nothing about the qeuue did change
+      [condition unlock];
+      return nil;
+    }
+  id obj =
+    [[(id<NSObject>)_items[_tail % _capacity] retain] autorelease];
+  [condition unlock];
+  return obj;
+}
+
+
+- (void*) peek
+{
+  if (condition != nil)
+    {
+      return [self _cooperatingPeek];
+    }
+  if (_head - _tail == 0)
+    {
+      return NULL;
+    }
+  return _items[_tail % _capacity];
+}
+
+- (NSObject*) peekObject
+{
+  if (condition != nil)
+    {
+      return [self _cooperatingPeekObject];
+    }
+  if (_head - _tail == 0)
+    {
+      return nil;
+    }
+  return [[(id<NSObject>)_items[_tail % _capacity] retain] autorelease];
+}
+
 - (unsigned) _cooperatingPut: (void**)buf
 		       count: (unsigned)count
 		 shouldBlock: (BOOL)block
@@ -274,6 +365,83 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
   return index;
 }
 
+- (void) putAll: (void**)buf count: (unsigned)count shouldRetain: (BOOL)rtn
+{
+  NSTimeInterval	ti;
+  unsigned		index;
+  BOOL			wasEmpty;
+
+  NSAssert(nil != condition, NSGenericException);
+  NSAssert(count <= _capacity, NSInvalidArgumentException);
+
+  [condition lock];
+  if (_head - _tail < count)
+    {
+      if (_head - _tail == _capacity)
+        {
+          _putTryFailure++;
+          fullCount++;
+        }
+
+      START
+      if (0 == timeout)
+	{
+	  while (_head - _tail < count)
+	    {
+	      [condition wait];
+	    }
+	}
+      else
+	{
+	  NSDate	*d;
+
+	  d = [[NSDateClass alloc]
+	    initWithTimeIntervalSinceNow: timeout / 1000.0f];
+	  while (_head - _tail < count)
+	    {
+	      if (NO == [condition waitUntilDate: d])
+		{
+		  [d release];
+		  ENDPUT
+		  [condition broadcast];
+		  [condition unlock];
+		  [NSException raise: NSGenericException
+			      format: @"Timeout waiting for space in FIFO"];
+		}
+	    }
+	  [d release];
+	}
+      ENDPUT
+    }
+  else
+    {
+      _putTrySuccess++;
+    }
+
+  if (_head - _tail == 0)
+    {
+      wasEmpty = YES;
+    }
+  else
+    {
+      wasEmpty = NO;
+    }
+  for (index = 0; index < count; index++)
+    {
+      _items[_head % _capacity] = buf[index];
+      _head++;
+      if (YES == rtn)
+        {
+          RETAIN((NSObject*)buf[index]);
+        }
+    }
+  if (YES == wasEmpty)
+    {
+      [condition broadcast];
+    }
+  [condition unlock];
+}
+
 - (void) dealloc
 {
   [classLock lock];
@@ -323,21 +491,28 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
     fullCount];
 }
 
-- (unsigned) get: (void**)buf count: (unsigned)count shouldBlock: (BOOL)block
+- (unsigned) get: (void**)buf
+           count: (unsigned)count
+     shouldBlock: (BOOL)block
+          before: (NSDate*)date
+
 {
   unsigned		index;
   NSTimeInterval	ti;
   NSTimeInterval	sum;
+  NSTimeInterval        waitLength;
   uint32_t		old;
-  uint32_t		fib;
-  
+  uint32_t		fib; 
   if (0 == count) return 0;
 
   if (nil != condition)
     {
-      return [self _cooperatingGet: buf count: count shouldBlock: block];
+      return [self _cooperatingGet: buf 
+                             count: count 
+                       shouldBlock: block
+                            before: date];
     }
-
+  waitLength = [date timeIntervalSinceNow];
   if (nil == getThread)
     {
       getThread = [NSThread currentThread];
@@ -374,6 +549,11 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
 	  [NSException raise: NSGenericException
 		      format: @"Timeout waiting for new data in FIFO"];
 	}
+      else if (date != nil && sum > waitLength)
+        {
+          ENDGET
+          return 0;
+        }
       tmp = fib + old;
       old = fib;
       fib = tmp;
@@ -394,6 +574,43 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
   return index;
 }
 
+
+- (unsigned) get: (void**)buf count: (unsigned)count shouldBlock: (BOOL)block
+{
+  return [self get: buf count: count shouldBlock: block before: nil];
+}
+
+
+
+- (unsigned) getObjects: (NSObject**)buf
+                  count: (unsigned)count
+            shouldBlock: (BOOL)block
+                 before: (NSDate*)date
+{
+  unsigned      result;
+  unsigned      index;
+
+  index = result = [self get: (void**)buf
+                       count: count 
+                 shouldBlock: block 
+                      before: date];
+  while (index-- > 0)
+    {
+      [buf[index] autorelease];
+    }
+  return result;
+}
+
+- (unsigned) getObjects: (NSObject**)buf
+                  count: (unsigned)count
+            shouldBlock: (BOOL)block
+{
+  return [self getObjects: buf
+                    count: count
+              shouldBlock: block
+                   before: nil];
+}
+
 - (void*) get
 {
   void	*item = 0;
@@ -403,6 +620,24 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
   return item;
 }
 
+- (NSObject*) getObject
+{
+  void	*item = 0;
+
+  while (0 == [self get: &item count: 1 shouldBlock: YES])
+    ;
+  return [(NSObject*)item autorelease];
+}
+
+- (NSObject*) getObjectRetained NS_RETURNS_RETAINED
+{
+  void	*item = 0;
+
+  while (0 == [self get: &item count: 1 shouldBlock: YES])
+    ;
+  return (NSObject*)item;
+}
+
 - (id) initWithCapacity: (uint32_t)c
 	    granularity: (uint16_t)g
 		timeout: (uint16_t)t
@@ -411,7 +646,7 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
 	     boundaries: (NSArray*)a
 		   name: (NSString*)n
 {
-  if (c < 1 || c > 1000000)
+  if (c < 1 || c > 100000000)
     {
       [self release];
       return nil;
@@ -486,6 +721,7 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
 {
   NSUserDefaults	*defs = [NSUserDefaults standardUserDefaults];
   NSString		*key;
+  NSInteger             i;
   uint16_t		g;
   uint16_t		t;
   BOOL			mc;
@@ -493,19 +729,27 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
   NSArray		*b;
 
   key = [NSString stringWithFormat: @"GSFIFOCapacity%@", n];
-  if ([defs integerForKey: key] > 0)
+  if (nil == [defs objectForKey: key]) key = @"GSFIFOCapacity";
+  i = [defs integerForKey: key];
+  if (i > 0)
     {
-      c = [defs integerForKey: key];
+      c = i;
     }
+
   key = [NSString stringWithFormat: @"GSFIFOGranularity%@", n];
+  if (nil == [defs objectForKey: key]) key = @"GSFIFOGranularity";
   g = [defs integerForKey: key];
   key = [NSString stringWithFormat: @"GSFIFOTimeout%@", n];
+  if (nil == [defs objectForKey: key]) key = @"GSFIFOTimeout";
   t = [defs integerForKey: key];
   key = [NSString stringWithFormat: @"GSFIFOSingleConsumer%@", n];
+  if (nil == [defs objectForKey: key]) key = @"GSFIFOSingleConsumer";
   mc = (YES == [defs boolForKey: key]) ? NO : YES;
   key = [NSString stringWithFormat: @"GSFIFOSingleProducer%@", n];
+  if (nil == [defs objectForKey: key]) key = @"GSFIFOSingleProducer";
   mp = (YES == [defs boolForKey: key]) ? NO : YES;
   key = [NSString stringWithFormat: @"GSFIFOBoundaries%@", n];
+  if (nil == [defs objectForKey: key]) key = @"GSFIFOBoundaries";
   b = [defs arrayForKey: key];
 
   return [self initWithCapacity: c
@@ -596,9 +840,47 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
   return index;
 }
 
+- (unsigned) putObjects: (NSObject**)buf
+                  count: (unsigned)count
+            shouldBlock: (BOOL)block
+{
+  unsigned      result;
+  unsigned      index;
+
+  /* NB we must retain objects *before* putting them in the FIFO since
+   * another thread may grab them immediately.
+   * That means, if we fail to put all of the objects, we must release
+   * any that were left over.
+   */
+  for (index = 0; index < count; index++)
+    {
+      [buf[index] retain];
+    }
+  result = [self put: (void**)buf count: count shouldBlock: block];
+  while (count-- > result)
+    {
+      [buf[count] release];
+    }
+  return result;
+}
+
+
 - (void) put: (void*)item
 {
-  while (0 == [self put: &item count: 1 shouldBlock: YES])
+  while (0 == [self put: (void**)&item count: 1 shouldBlock: YES])
+    ;
+}
+
+- (void) putObject: (NSObject*)item
+{
+  [item retain];
+  while (0 == [self put: (void**)&item count: 1 shouldBlock: YES])
+    ;
+}
+
+- (void) putObjectConsumed: (NSObject*) NS_CONSUMED item
+{
+  while (0 == [self put: (void**)&item count: 1 shouldBlock: YES])
     ;
 }
 
@@ -747,6 +1029,14 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
   return item;
 }
 
+- (NSObject*) tryGetObject
+{
+  void	*item = nil;
+  
+  [self get: &item count: 1 shouldBlock: NO];
+  return [(NSObject*)item autorelease];
+}
+
 - (BOOL) tryPut: (void*)item
 {
   if (1 == [self put: &item count: 1 shouldBlock: NO])
@@ -756,5 +1046,32 @@ stats(NSTimeInterval ti, uint32_t max, NSTimeInterval *bounds, uint64_t *bands)
   return NO;
 }
 
+- (BOOL) tryPutObject: (NSObject*)item
+{
+  [item retain];
+  if (1 == [self put: (void**)&item count: 1 shouldBlock: NO])
+    {
+      return YES;
+    }
+  [item release];
+  return NO;
+}
+
+- (NSUInteger)sizeInBytesExcluding: (NSHashTable*)excluding
+{
+  NSUInteger size = 0;
+  if (0 == (size = [super sizeInBytesExcluding: excluding]))
+    {
+      return 0;
+    }
+  return size
+   + (_capacity * sizeof(void)) // item storage
+   + (boundsCount * sizeof(NSTimeInterval)) // boundaries
+   + (2 * (boundsCount + 1) * sizeof(uint64_t)) // get and put counts
+   + [condition sizeInBytesExcluding: excluding]
+   + [name sizeInBytesExcluding: excluding]
+   + [putThread sizeInBytesExcluding: excluding]
+   + [getThread sizeInBytesExcluding: excluding];
+}
 @end
 
