@@ -22,7 +22,7 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
-   $Date: 2011-09-02 10:33:25 +0100 (Fri, 02 Sep 2011) $ $Revision: 33807 $
+   $Date$ $Revision$
    */ 
 
 #import	<Foundation/NSArray.h>
@@ -500,7 +500,11 @@ typedef struct {
 
 - (void) add: (unsigned)count
 {
-  NSAssert(my->supportDurations == NO, @"configured for durations");
+  if (NO != my->supportDurations)
+    {
+      [NSException raise: NSInternalInconsistencyException
+                  format: @"-add: called when set for durations"];
+    }
   if (my->numberOfPeriods == 0)
     {
       cseconds[0].cnt += count; // Total
@@ -514,7 +518,11 @@ typedef struct {
 
 - (void) add: (unsigned)count duration: (NSTimeInterval)length
 {
-  NSAssert(my->supportDurations == YES, @"not configured for durations");
+  if (YES != my->supportDurations)
+    {
+      [NSException raise: NSInternalInconsistencyException
+                  format: @"-add:duration: called when not set for durations"];
+    }
 
   if (count > 0)
     {
@@ -567,7 +575,12 @@ typedef struct {
   unsigned      from;
   unsigned      to;
 
-  NSAssert(my->supportDurations == YES, @"not configured for durations");
+  if (YES != my->supportDurations)
+    {
+      [NSException raise: NSInternalInconsistencyException
+                  format: @"-addDuration: called when not set for durations"];
+    }
+
   if (my->numberOfPeriods == 0)
     {
       from = 0; // Total
@@ -643,6 +656,34 @@ typedef struct {
   [super dealloc];
 }
 
+static void 
+appendCountInfo(CountInfo *info, NSMutableString *m, NSTimeInterval base)
+{
+  NSDate	*d = [NSDate alloc];
+
+  d = [d initWithTimeIntervalSinceReferenceDate: info->tick + base];
+  [m appendFormat: @"%u, %@\n", info->cnt, d];
+  RELEASE(d);
+}
+
+static void 
+appendDurationInfo(DurationInfo *info, NSMutableString *m, NSTimeInterval base)
+{
+  NSDate	*d = [NSDate alloc];
+
+  d = [d initWithTimeIntervalSinceReferenceDate: info->tick + base];
+  if (info->cnt)
+    {
+      [m appendFormat: @"%u, %g, %g, %g, %@\n",
+	info->cnt, info->max, info->min, info->sum, d];
+    }
+  else
+    {
+      [m appendFormat: @"0, -, -, -, %@\n", d];
+    }
+  RELEASE(d);
+}
+
 - (NSString*) description
 {
   NSAutoreleasePool     *pool = [NSAutoreleasePool new];
@@ -689,16 +730,12 @@ typedef struct {
 		  tick = 0;
 		  for (i = 0; i < my->second; i++)
 		    {
-		      DurationInfo		*info = &dseconds[i];
-		      NSTimeInterval	ti = info->tick + baseTime;
+		      DurationInfo	*info = &dseconds[i];
 
 		      if (info->tick != tick)
 			{
 			  tick = info->tick;
-			  [m appendFormat: @"%u, %g, %g, %g, %@\n",
-			    info->cnt, info->max, info->min, info->sum,
-			    [NSDate dateWithTimeIntervalSinceReferenceDate:
-			      ti]];
+			  appendDurationInfo(info, m, baseTime);
 			}
 		    }
 		}
@@ -709,16 +746,12 @@ typedef struct {
 		  tick = 0;
 		  for (i = 0; i < my->minute; i++)
 		    {
-		      DurationInfo		*info = &dminutes[i];
-		      NSTimeInterval	ti = info->tick + baseTime;
+		      DurationInfo	*info = &dminutes[i];
 
 		      if (info->tick != tick)
 			{
 			  tick = info->tick;
-			  [m appendFormat: @"%u, %g, %g, %g, %@\n",
-			    info->cnt, info->max, info->min, info->sum,
-			    [NSDate dateWithTimeIntervalSinceReferenceDate:
-			      ti]];
+			  appendDurationInfo(info, m, baseTime);
 			}
 		    }
 		}
@@ -731,32 +764,24 @@ typedef struct {
                    */
 		  for (i = my->period; i < my->numberOfPeriods; i++)
 		    {
-		      DurationInfo		*info = &dperiods[i];
-		      NSTimeInterval	ti = info->tick + baseTime;
+		      DurationInfo	*info = &dperiods[i];
 
 		      if (info->tick != tick)
 			{
 			  tick = info->tick;
-			  [m appendFormat: @"%u, %g, %g, %g, %@\n",
-			    info->cnt, info->max, info->min, info->sum,
-			    [NSDate dateWithTimeIntervalSinceReferenceDate:
-			      ti]];
+			  appendDurationInfo(info, m, baseTime);
 			}
 		    }
                   /* Periods from current cycle
                    */
 		  for (i = 0; i < my->period; i++)
 		    {
-		      DurationInfo		*info = &dperiods[i];
-		      NSTimeInterval	ti = info->tick + baseTime;
+		      DurationInfo	*info = &dperiods[i];
 
 		      if (info->tick != tick)
 			{
 			  tick = info->tick;
-			  [m appendFormat: @"%u, %g, %g, %g, %@\n",
-			    info->cnt, info->max, info->min, info->sum,
-			    [NSDate dateWithTimeIntervalSinceReferenceDate:
-			      ti]];
+			  appendDurationInfo(info, m, baseTime);
 			}
 		    }
 		}
@@ -770,14 +795,11 @@ typedef struct {
 		  for (i = 0; i < my->second; i++)
 		    {
 		      CountInfo		*info = &cseconds[i];
-		      NSTimeInterval	ti = info->tick + baseTime;
 
 		      if (info->tick != tick)
 			{
 			  tick = info->tick;
-			  [m appendFormat: @"%u, %@\n", info->cnt,
-			    [NSDate dateWithTimeIntervalSinceReferenceDate:
-			      ti]];
+			  appendCountInfo(info, m, baseTime);
 			}
 		    }
 		}
@@ -789,14 +811,11 @@ typedef struct {
 		  for (i = 0; i < my->minute; i++)
 		    {
 		      CountInfo		*info = &cminutes[i];
-		      NSTimeInterval	ti = info->tick + baseTime;
 
 		      if (info->tick != tick)
 			{
 			  tick = info->tick;
-			  [m appendFormat: @"%u, %@\n", info->cnt,
-			    [NSDate dateWithTimeIntervalSinceReferenceDate:
-			      ti]];
+			  appendCountInfo(info, m, baseTime);
 			}
 		    }
 		}
@@ -810,14 +829,11 @@ typedef struct {
 		  for (i = my->period; i < my->numberOfPeriods; i++)
                     {
 		      CountInfo		*info = &cperiods[i];
-		      NSTimeInterval	ti = info->tick + baseTime;
 
 		      if (info->tick != tick)
 			{
 			  tick = info->tick;
-			  [m appendFormat: @"%u, %@\n", info->cnt,
-			    [NSDate dateWithTimeIntervalSinceReferenceDate:
-			      ti]];
+			  appendCountInfo(info, m, baseTime);
 			}
                     }
                   /* Periods from current cycle
@@ -825,14 +841,11 @@ typedef struct {
 		  for (i = 0; i < my->period; i++)
 		    {
 		      CountInfo		*info = &cperiods[i];
-		      NSTimeInterval	ti = info->tick + baseTime;
 
 		      if (info->tick != tick)
 			{
 			  tick = info->tick;
-			  [m appendFormat: @"%u, %@\n", info->cnt,
-			    [NSDate dateWithTimeIntervalSinceReferenceDate:
-			      ti]];
+			  appendCountInfo(info, m, baseTime);
 			}
 		    }
 		}
@@ -848,9 +861,12 @@ typedef struct {
 {
   if (my->started > 0.0)
     {
-      [self addDuration: (*tiImp)(NSDateClass, tiSel) - my->started];
+      NSTimeInterval    ti;
+
+      ti = (*tiImp)(NSDateClass, tiSel) - my->started;
       my->event = nil;
       my->started = 0.0;
+      [self addDuration: ti];
     }
 }
 
@@ -1019,12 +1035,20 @@ typedef struct {
 
 - (void) startDuration: (NSString*)name
 {
-  NSAssert(my->supportDurations == YES && my->started == 0.0,
-    NSInternalInconsistencyException);
+  if (NO == my->supportDurations)
+    {
+      [NSException raise: NSInternalInconsistencyException
+        format: @"-startDuration: for '%@' when not set for durations", name];
+    }
+  if (0.0 != my->started)
+    {
+      [NSException raise: NSInternalInconsistencyException
+        format: @"-startDuration: for '%@' when already started", name];
+    }
   if (my->event != nil)
     {
       [NSException raise: NSInternalInconsistencyException
-		  format: @"-startDuration: for '%@' nested inside '%@'",
+        format: @"-startDuration: for '%@' nested inside '%@'",
 	my->event, name];
     }
   my->started = (*tiImp)(NSDateClass, tiSel);
